package define

// AutoUpdateLabel denotes the container/pod label key to specify auto-update
// policies in container labels.
const AutoUpdateLabel = "io.containers.autoupdate"

// AutoUpdateAuthfileLabel denotes the container label key to specify authfile
// in container labels.
const AutoUpdateAuthfileLabel = "io.containers.autoupdate.authfile"
