// Code generated by go-swagger; DO NOT EDIT.

package network

// This file was generated by the swagger tool.
// Editing this file might prove futile when you re-run the swagger generate command

// Summary Network list response item
//
// swagger:model Summary
type Summary struct {
	Network
}
